import urllib.request, urllib.parse, urllib.error
import http.client
import string
import base64
import json
import jetpack.util
import jetpack.config
import urllib.parse
import logging
import math
from typing import Optional, Union, Any, Dict


import ssl
from copy import deepcopy
_unverified_context = ssl._create_unverified_context()
_unverified_context.options |= ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 | ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1 | ssl.OP_NO_COMPRESSION


logger = logging.getLogger()


def set(name="execute", corecount=0, instancecount=None):
    """ Quick set an array to a target core count """
    nodearray = {
        'Name': name
    }
    if instancecount:
        nodearray['TargetCount'] = instancecount
    else:
        nodearray['TargetCoreCount'] = corecount

    arrays = [nodearray]
    update(arrays)


def update(arrays: Union[list, dict], request_set: Optional[str] = None, expiration_interval: Optional[int] = None, config: Optional[dict] = None):
    """ Set several arrays in bulk """
    configuration = jetpack.util.parse_config(config)

    cluster_name = configuration['identity']['cluster_name']
    cluster_start_count = configuration['identity']['cluster_start_count']

    url = "/cloud/api/autoscale/%s" % urllib.parse.quote(cluster_name)

    params: Dict[str, Any] = {}
    if expiration_interval:
        params['expiration_interval'] = expiration_interval
    if request_set:
        params['request_set'] = request_set
    params['start_count'] = cluster_start_count

    # Ensure that the body is a list of (of dictionaries)
    if not isinstance(arrays, list):
        body = [arrays]
    body = arrays  # type: ignore # mypy loses it's mind assigning union types

    json_body = json.dumps(body, sort_keys=True)
    logger.info("Making autoscale request to %s: %s" % (url, json_body))

    response = jetpack.util.query_cyclecloud(url, body=json_body, params=params, method="POST")

    if response.status != 200:
        logger.error("An error occurred making autoscale request to %s: %s (%s)" % (url, response.reason, response.status))
        raise Exception(response.reason)


def scale_by_jobs(jobs, definitions=None, default_average_runtime=60, default_nodearray="execute", request_set=None, expiration_interval=None, config=None):
    """ Map jobs by nodearray and total runtime hours 
    
    jobs is a list of dictionaries describing jobs keys are:
        - name: the name of the job (required)
        - nodearray: the name of the nodearray the job belongs to (if None, uses `default_nodearray`)
        - average_runtime: the average runtime of the job in minutes (if None, uses `default_average_runtime`)
        - request_cpus: how many cpus this job is going to use
    """
    if definitions is None:
        definitions = []
    runtime_by_array = {}
    grouped_runtime_by_array = {}

    for job in jobs:
        array_name = job.get('nodearray', default_nodearray)
        if array_name is None:
            array_name = default_nodearray
        array_name = array_name.lower()

        average_runtime = job.get('average_runtime', default_average_runtime)
        if average_runtime is None:
            average_runtime = default_average_runtime
        average_runtime = min(average_runtime, 60.0)

        request_cpus = job.get('request_cpus', 1)
        if request_cpus is None:
            request_cpus = 1

        grouped = job.get('grouped', False)

        if grouped:
            grouped_runtime_by_array[array_name] = grouped_runtime_by_array.get(array_name, 0) + average_runtime * request_cpus
        else:
            runtime_by_array[array_name] = runtime_by_array.get(array_name, 0) + average_runtime * request_cpus
    
    arrays = []

    # map definitions by "name" (lowercase)
    definitions_by_name = {}
    for definition in definitions:
        for key in definition.keys():
            if key.lower() == "name":
                definitions_by_name[definition[key]] = definition

    for name, runtime in runtime_by_array.items():
        array = deepcopy(definitions_by_name.get(name.lower(), {}))

        array.update({
            'Name': name,
            'TargetCoreCount': int(math.ceil(runtime / 60.0))
        })
        arrays.append(array)

    for name, runtime in grouped_runtime_by_array.items():
        grouped_name = '%s:grouped' % (name)
        array = definitions_by_name.get(grouped_name.lower())
        
        # fall back on non-grouped name for backwards compatibility
        if not array:
            array = definitions_by_name.get(name.lower(), {})
        array = deepcopy(array)

        array.update({
            'TargetCoreCount': int(math.ceil(runtime / 60.0)),
            'Extends': name,
            'Dynamic': True,
            'Azure': {'SingleScaleset': True},
            'Name': grouped_name
        })

        arrays.append(array)
    
    update(arrays, request_set=request_set, expiration_interval=expiration_interval, config=config)
