import subprocess
import click # type: ignore
from jetpack import util
import traceback
import os
import json
import logging
import logging.handlers
import logging.config

from jetpack.util import UserError

NESTED_INDICATOR = "<nested>"


def _handle_exception():
    logger = logging.getLogger()
    logger.error(traceback.format_exc())
    raise click.ClickException("An unknown error occurred, please see jetpack.log for details.")


# Note: Not sure the params here, taken from docs
def _print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return

    try:
        version = util.version()
    except Exception as e:
        raise click.ClickException(str(e))

    click.echo("%s" % version)
    ctx.exit()


def _format_dict(_dict, _val=None):
    results = {}
    for _k in list(_dict.keys()):
        subdict = _dict[_k]
        if isinstance(subdict, dict):
            attrib = NESTED_INDICATOR
        else:
            attrib = subdict
        if _val:
            formatted_key = "%s.%s" % (_val, _k)
        else:
            formatted_key = "%s" % _k
        results[formatted_key] = attrib
    formatted_keys = ""
    _configs = sorted(results.keys())
    max_key_length = len(max(_configs, key=len))
    for _config in _configs:
        formatted_key = _config.ljust(max_key_length)
        formatted_keys += "\n  %s = %s" % (formatted_key, results[_config])
    return formatted_keys


@click.group()
@click.option('--loglevel', type=click.Choice(["debug", "info", "warn"]), default='info')
@click.option('-c', '--config', type=click.Path(), help="Jetpack configuration file")
@click.option('-v', '--version', is_flag=True, callback=_print_version, expose_value=False, is_eager=True, help="Version information for this program")
@click.pass_context
def jetpack_group(ctx, loglevel, config=None):
    ctx.obj = {}
    ctx.obj['config'] = util.parse_config(config)

    util.setup_logging(loglevel)


@jetpack_group.command(name='keepalive', short_help="Delays the healthcheck daemon")
@click.argument('duration', required=False)
def keepalive_command(duration=None):
    """
    Delays the healthcheck daemon from terminating the system for a specified period of time.

    DURATION defaults to 1 hour if not specified

    Valid values for DURATION are of the format n{m,h,d} where n is an integer and {m,h,d} are time units.
    "h" is the default time unit if no time unit specified

    time units:
    m - minutes
    h - hours
    d - days

    \b
    Examples:
    $ jetpack keepalive            # system will be kept alive for 1 hour
    $ jetpack keepalive forever    # healthservice will be disabled and system will stay up indefinitely
    $ jetpack keepalive 2d         # system will be kept alive for 2 days
    """
    from jetpack import keepalive
    try:
        msg = keepalive.execute(duration)
        click.echo(msg)
    except keepalive.KeepaliveError as e:
        raise click.ClickException(str(e))
    except click.ClickException as e:
        raise e
    except Exception:
        _handle_exception()


@jetpack_group.command(name='users')
@click.option('--owner', 'is_owner', is_flag=True, help="Output the cluster owner")
@click.option('--json', 'as_json', is_flag=True, help="Output formatted as JSON")
def users_command(is_owner, as_json):
    from jetpack import users

    results = []
    if is_owner:
        results = users.owner()
        if results is None:
            raise click.ClickException("No cluster owner found")
    else:
        results = users.query()

    if as_json:
        click.echo(json.dumps(results, sort_keys=True, indent=4))
    else:
        if not isinstance(results, list):
            results = [results]
        for user in results:
            click.echo("Username: %s" % user.get('name'))
            click.echo("Full Name: %s" % user.get('fullName', ''))
            click.echo("UID: %s" % user.get('uid', ''))
            click.echo("Is Admin: %s" % user.get('isAdmin', False))
            click.echo("Is Owner: %s" % user.get('isOwner', False))
            click.echo()


@jetpack_group.command(name='config')
@click.argument('value', required=False)
@click.argument('default', required=False)
@click.option('-j', '--json-file', type=click.Path(), default=None, help="Non-default node configuration file to use")
@click.option('--json', 'as_json', is_flag=True, help="Output data in serial json format")
def config_command(value, default, json_file, as_json):
    """Gets a configuration value"""

    from jetpack import config

    # Find the deepest valid configuration
    if value:
        properties = value.split('.')
    else:
        properties = []
    while True:
        _key = '.'.join(properties)
        if _key == "":
            _key = None
        try:
            v = config.get(_key, default, configuration_file=json_file)
        except config.ConfigError as e:
            if _key is None or as_json:
                raise click.ClickException(str(e))
            else:
                properties.pop()
                continue
        break

    if as_json:
        if not(isinstance(v, dict)):
            # Force into valid dict for json module, if only a single value.
            v = {value: v}
        if _key != value:
            raise click.ClickException("The value '%s' was not found in the node configuration." % value)
        click.echo(json.dumps(v, indent=4))
        return

    err = False
    MSG = []
    if isinstance(v, dict):
        output = _format_dict(v, _val=_key)
        err = True
    else:
        output = str(v)
    if value != _key:
        MSG.append("The value '%s' was not found in the node configuration." % value)
        if _key:
            MSG.append("  Reporting results for '%s'" % _key)
        err = True
    if _key is None:
        MSG.append("  Reporting top level options.")
        err = True
    if err:
        MSG.append(output)
        raise click.ClickException('\n'.join(MSG) + "\n")
    else:
        click.echo(output)


@jetpack_group.command(name='converge')
@click.option('--mode', type=click.Choice(['install', 'maintenance']))
@click.option('--cluster-init', 'clusterinit', is_flag=True, help="Only run cluster-init scripts")
@click.option('--debug', is_flag=True, help="Run in debug mode")
@click.option('--no-sync', is_flag=True, help="Do not sync Chef repositories before executing converge")
def converge_command(debug, no_sync, clusterinit, mode):
    """Triggers a converge on the current node"""
    try: 
        # Note: This is a temporary hack to allow users to run cluster-init from the CLI
        # In the future this will be replaced with a "jetpack cluster_init" command
        # or something similar
        if clusterinit:
            from jetpack import cluster_init
            try:
                cluster_init.process_scripts()
            except cluster_init.ClusterInitError as e:
                raise click.ClickException(str(e))
            except click.ClickException as e:
                raise e
            except Exception:
                _handle_exception()

        else:
            from jetpack import converge
            try:
                converge.execute(debug=debug, no_sync=no_sync, mode=mode)
            except ImportError:
                # This happens when there are missing dependencies in the omnibus build
                _handle_exception()
            except converge.ConvergeError as e:
                raise click.ClickException(str(e))
            except click.ClickException as e:
                raise e
            except Exception:
                _handle_exception()
            click.echo("Node converged successfully")
    finally:
        if not util.is_windows():
            logs_path = os.path.join(util.jetpack_home(), 'logs')
            subprocess.check_call(f"chown -R root:cyclecloud {logs_path}", shell=True)


@jetpack_group.command(name='log')
@click.argument('message')
@click.option('-l', '--level', default='info', type=click.Choice(['info', 'warn', 'error']), help="Type of log message you wish to send. Default: info")
@click.option('-p', '--priority', type=click.Choice(['low', 'medium', 'high']), help="""
Priority of message to send to CycleCloud. Default depends on the level chosen. Info messages get a priority of 'medium', warn messages 'medium', and error messages 'high'""")
@click.pass_context
def log_command(ctx, message, level, priority=None):
    """Sends log message to CycleCloud"""
    from jetpack import messaging
    try:
        messaging.log(message, level, priority)
    except messaging.LogError as e:
        raise click.ClickException(str(e))
    except click.ClickException as e:
        raise e
    except Exception:
        _handle_exception()


@jetpack_group.command(name='send')
@click.option('-m', '--message', help="The message to send")
@click.option('-f', '--file', type=click.Path(), help="The file to send.")
@click.option('-r', '--routing-key', default=None, help="The amqp routing-key associated with the message")
@click.option('-e', '--exchange', default=None, help="The amqp exchange associated with the message")
@click.pass_context
def send_command(ctx, message, file, routing_key, exchange):
    """ Sends a message to CycleCloud """
    from jetpack import amqp
    try:
        amqp.send(message=message, file=file, routing_key=routing_key, exchange=exchange, config=ctx.obj['config'])
    except amqp.SendError as e:
        raise click.ClickException(str(e))
    except click.ClickException as e:
        raise e
    except Exception:
        _handle_exception()


@jetpack_group.command(name='test')
@click.option('-s', '--stream-output', is_flag=True, default=False, help="Stream output of test result")
@click.option('--send-report', is_flag=True, default=False, help="Send test report to CycleCloud")
@click.pass_context
def test_command(ctx, stream_output, send_report):
    """ Runs python style unittests on the node """
    from jetpack import cluster_init

    try:
        report = cluster_init.run_tests(capture=not(stream_output), send_report=send_report)

        # The tests failed, bail
        if report.num_failed > 0:
            raise click.ClickException("One or more tests failed, check %s for details" % cluster_init._get_results_filepath())

    except cluster_init.JetpackTestException as e:
        raise click.ClickException(str(e))
    except click.ClickException as e:
        raise e
    except Exception:
        _handle_exception()


@jetpack_group.command(name='autoscale')
@click.option('-f', '--file', type=click.Path(exists=True), help="JSON formatted file containing a list of nodearray definitions.")
@click.option('-n', '--name', default='execute', help="The name of a nodearray to autoscale.")
@click.option('-c', '--corecount', type=click.INT, default=None, help="The number of cores to autoscale the nodearray to.")
@click.option('-i', '--instancecount', type=click.INT, default=None, help="The total number of instances to autoscale the nodearray to.")
@click.pass_context
def autoscale_command(ctx, file, name, corecount, instancecount):
    """ Autoscales the cluster the node is part of """
    from jetpack import autoscale
    if file:
        try:
            arrays = json.load(open(file))
            autoscale.update(arrays)
        except ValueError:
            raise click.ClickException("The file specified does not appear to be valid json")

    elif name is not None and (corecount is not None or instancecount is not None):
        autoscale.set(name, corecount, instancecount)
    else:
        raise click.ClickException("Please specify either a file, or an array name and target count - see help for details.")


@jetpack_group.command(name='shutdown')
@click.option('--idle', is_flag=True, help="The node is idle and can be removed")
@click.option('--unhealthy', is_flag=True, help="The node is unhealthy and has to be removed")
@click.option('--deallocate', is_flag=True, help="Request that the node be deallocated")
@click.option('--terminate', is_flag=True, help="Request that the node be terminated")
@click.pass_context
def shutdown_command(ctx, idle, unhealthy, deallocate, terminate):
    """ terminates the node """
    from jetpack import shutdown
    try:
        if not idle and not unhealthy:
            click.confirm('Are you sure you want to shutdown this node?', abort=True)

        shutdown.execute(idle=idle, unhealthy=unhealthy, deallocate=deallocate, terminate=terminate)
    except click.Abort:
        return
    except shutdown.ShutdownError as e:
        raise click.ClickException(str(e))
    except click.ClickException as e:
        raise e
    except KeyboardInterrupt:
        pass
    except Exception:
        click.echo(traceback.format_exc())
        _handle_exception()


@jetpack_group.command(name='run_on_shutdown', short_help="Add a script to be called prior to node termination")
@click.argument('script_path', required=False, type=click.Path(exists=True))
@click.pass_context
def run_on_shutdown(ctx, script_path):
    """
    Registers a script to be executed prior to termination (and with best-effort on preemption).

    Shutdown actions are executed in the reverse order they are registered to allow unwinding startup steps.
    """
    from jetpack import shutdown_actions
    try:
        if not os.path.exists(script_path):
            raise click.ClickException("The script '%s' does not exist" % script_path)
        if not os.access(script_path, os.X_OK):
            raise click.ClickException("The script '%s' is not executable" % script_path)

        shutdown_actions.register_shutdown_script(script_path)
    except click.ClickException as e:
        raise e
    except Exception:
        _handle_exception()



@jetpack_group.command(name='download')
@click.argument('path', type=click.Path())
@click.argument('dst', required=False, default=".")
@click.option('--project', envvar='CYCLECLOUD_PROJECT_NAME', help="The project the blob is in", required=True)
@click.option('--spec', envvar='CYCLECLOUD_SPEC_NAME', help="the project spec to use")
@click.pass_context
def download_command(ctx, path, dst, project, spec):
    """Download a project file from cloud storage"""
    from jetpack import cluster_init
    try:
        cluster_init.download_blob(path, dst, project)
    except (UserError, cluster_init.ClusterInitError) as e:
        raise click.ClickException(str(e))
    except click.ClickException as e:
        raise e
    except Exception:
        _handle_exception()


@jetpack_group.command(name='report_issue')
@click.argument('path', required=False, type=click.Path(exists=True))
@click.option('--upload/--no-upload', default=True, help="Upload local logs to Azure Storage")
@click.option('--sign/--no-sign', default=True, help="Sign the blob in Azure Storage for temporary read only access")
@click.option('--client-id', help="An optional managed identity to use to access the locker, required if the locker's identity does not have write permission")
@click.pass_context
def report_issue_command(ctx: click.Context, path: str, upload: bool, sign: bool, client_id: str) -> None:
    """ Upload logs to Azure Storage. """
    from jetpack import report_issue
    try:
        log_path: str = report_issue.upload_logs(path=path, upload=upload, sign=sign, client_id=client_id)
        click.echo("Logs can be found at: %s" % log_path)

        # Attempt to parse the path (it may be a URL with an expiration time that we should display)
        if log_path.startswith("https"):
            import urllib.parse
            r = urllib.parse.urlparse(log_path)
            data = urllib.parse.parse_qs(r.query)
            if 'se' in data:
                signed_expiry = data['se'][0]
                click.echo("Signed URL will expire on: %s" % signed_expiry)

    except report_issue.UploadLogsError as e:
        raise click.ClickException(str(e))
    except click.ClickException as e:
        raise e
    except Exception:
        _handle_exception()


if __name__ == "__main__":
    jetpack_group()
