import jetpack.util
from jetpack.cluster_init import pytest_wrapper
import logging
from jetpack.util import azcopy
import jetpack.converge.userdata
import jetpack.lockers
import http.client
import json
from subprocess import Popen, STDOUT
import os
import stat
import tempfile
import shutil
import subprocess
import tarfile
from jetpack import config as jetpack_config
import urllib.request
import urllib.parse
import urllib.error
import urllib.parse

from typing import List

from jetpack.util import JETPACK_USER_AGENT
from jetpack.util import UserError
import jetpack.config

logger = logging.getLogger()

# requests has some wonky logging which can be really spammy so just disable it for now
logging.getLogger("urllib3").setLevel(logging.CRITICAL)
logging.getLogger('requests').setLevel(logging.CRITICAL)


# Add a streaming logger so users see output on the CLI as well as jetpack.log
logger.addHandler(logging.StreamHandler())


class ClusterInitError(Exception):
    pass


class SourceNotFound(ClusterInitError):
    pass


# simple cache to avoid downloading the manifest multiple times per process.
# Currently the Chef part of the converge process spawns the jetpack process multiple times, so this
# cache does not remove all duplicate downloads, but it does eliminate a lot. For instance, on a PBS head node,
# this removes the number of manifest downloads from 14 to 5, out of a total of 21 azcopy downloads.
# Each azcopy invocation takes over 2 seconds, so this is a significant savings.
MANIFEST_DOWNLOADED_CACHE = {}


class ClusterInitSpec:
    def __init__(self, spec, config=None, load_from_manifest=False):
        """ Takes a dictionary definition of the spec """
        self.project_name = spec.get("project")
        self.project_version = spec.get("version")
        self.name = spec.get("project_spec")
        if None in [self.project_name, self.project_version, self.name]:
            raise ClusterInitError("Project name, spec and version must be specified")

        self.locker = spec.get('locker')
        self.config = config
        self.location = spec.get("location")
        if self.location:
            self.location = self.location.rstrip("/")
        self.run_list = None

        # Load any values we care about from the manifest
        if load_from_manifest:
            self._load_from_manifest()

        # Load the runlist from the passed in definition
        # If a spec was previously loaded from the manifest, this will override
        # it.
        user_defined_run_list = spec.get("run_list", None)
        if user_defined_run_list is not None:
            self.run_list = user_defined_run_list

    def _load_from_manifest(self):
        """ Downloads the manifest.json from the project locker and reads the spec configuration """
        
        # The location specified includes the spec in the path, we need to strip that off to get 
        # the manifest location
        src = self.location[:-len("/%s" % self.name)]
        src = "%s/manifest.json" % src 
        dst_dir = os.path.join(jetpack.util.jetpack_home(), 'config', 'projects', self.project_name)
        dst = os.path.join(dst_dir, 'manifest.json')
        
        if not os.path.isdir(dst_dir):
            os.makedirs(dst_dir)

        global MANIFEST_DOWNLOADED_CACHE
        # the manifest is per project, not per spec
        manifest_key = (self.project_name, self.project_version)
        if manifest_key not in MANIFEST_DOWNLOADED_CACHE:
            try:
                self._download(src, dst, verify_source=True)
            except SourceNotFound as e:
                # We should add a help_link here
                raise jetpack.util.UserError("Project download failed",
                    description="Failed to find project %s in locker at %s" % (self.project_name, src),
                    recommendation="Check that you have uploaded the project and correctly referenced it in your template",
                    help_link=None)

            MANIFEST_DOWNLOADED_CACHE[manifest_key] = True

        if os.path.isfile(dst):
            with open(dst, 'r') as f:
                manifest = json.load(f)

            specs = manifest.get('specs', {})

            self.run_list = specs.get(self.name, {}).get('run_list', None)

    def _download(self, src, dst, delete_removed=False, verify_source=False, directory=False):
        """ Download SRC -> DST using azcopy
   
        delete_removed: remove local files that don't exist on the remote
        verify_source: verify that a source file exists before taking any action
        directory: if true, operate on an entire directory instead of a single file
        """

        locker = jetpack.lockers.get(self.locker)

        # Perform an existence check on the source file prior to attempting a download.
        # Since we have seen issues in the past with DNS returning different results for the musl build compared to the azcopy command,
        # which can be very hard to debug, we just rely on azcopy reporting that the file does not exist.
        if verify_source and not azcopy.file_exists(src, locker):
            raise SourceNotFound(f"Source file {src} not found in locker {locker}")

        if directory is False and not os.path.isfile(dst):
            try:
                azcopy.copy(src, dst, src_locker=locker)
            except UserError:
                raise
            except Exception as e:
                raise ClusterInitError("Unable to copy %s to %s: %s" % (src, dst, e))
        elif directory is True:
            try:
                # AzCopy requires a trailing slash if the src and dst are directories
                # (src is expected to be a URL)
                if src[-1] != "/":
                    src = src + "/"
                if dst[-1] != os.path.sep:
                    dst = dst + os.path.sep

                azcopy.sync(src, dst, src_locker=locker, recursive=True, delete_destination=delete_removed)
            except UserError as e:
                raise e
            except Exception as e:
                raise ClusterInitError("Unable to sync %s to %s: %s" % (src, dst, e))
        else:
            try:
                azcopy.sync(src, dst, src_locker=locker, recursive=True, delete_destination=delete_removed)
            except UserError as e:
                raise e
            except Exception as e:
                raise ClusterInitError("Unable to sync %s to %s: %s" % (src, dst, e))

    def download_chef(self):
        """ Download the chef repositories to "root" """

        # Download files to the tarball cache/<project>/<spec>
        cache = os.path.join(jetpack.util.jetpack_home(), "system", "chef", "tarballs", self.project_name, self.name)
        if not os.path.isdir(cache):
            os.makedirs(cache)

        self._download("%s/chef" % self.location, cache, delete_removed=True, directory=True)

        # TODO: This is kinda crappy, should be a better way to handle this path
        disk_location = os.path.join('/opt', 'cycle', 'jetpack', 'system', 'chef', 'chef-repo')
        if jetpack.util.is_windows():
            disk_location = os.path.join('C:\\', 'cycle', 'jetpack', 'system', 'chef', 'chef-repo')
        
        _process_chef_tarball(cache, "site-cookbooks.tgz", disk_location, "cookbooks")
        _process_chef_tarball(cache, "roles.tgz", disk_location, "roles")
        _process_chef_tarball(cache, "data_bags.tgz", disk_location, "data_bags")

    def download_scripts(self, root=None):
        if root is None:
            root = jetpack.config.get('cluster_init.home_dir')
        
        src = "%s/cluster-init" % self.location
        dst = os.path.join(root, self.project_name, self.name)

        if not os.path.isdir(dst):
            os.makedirs(dst)
        self._download(src, dst, directory=True)

    @property
    def path(self):
        return os.path.join(_get_cluster_init_home(), self.project_name,
                            self.name)
    
    def test(self, capture=True):
        '''Execute integration tests for the current spec if present'''
        exitstatus, report = pytest_wrapper.run(self.path, self.project_name, self.name,
                                                capture=capture)
        return exitstatus, report
        
    def run(self, root=None, fail_on_error=None, allowed_executables=None, rerun=False):
        """ Run the ClusterInitSpec

        root: the root directory where files shoudl be run from
        force: should the scripts be run even if they've already completed before
        """
        logger.info("Running cluster-init scripts for project '%s v%s' spec '%s'" , self.project_name, self.project_version, self.name)
        if root is None:
            root = jetpack.config.get('cluster_init.home_dir')    
        if fail_on_error is None:
            fail_on_error = jetpack.config.get('cluster_init.fail_on_error')
        if allowed_executables is None:
            allowed_executables = jetpack.config.get('cluster_init.allowed_executables')

        run_dir, scripts_dir, log_dir = self._create_directories(root)

        for filename in sorted(os.listdir(scripts_dir)):
            _, ext = os.path.splitext(filename)
            if ext in allowed_executables:
                script = os.path.join(scripts_dir, filename)
                run_file = os.path.join(run_dir, filename + '.run')
                output_file = os.path.join(log_dir, filename + '.out')

                if rerun is False and os.path.isfile(run_file):
                    logger.info("Script %s has already run successfully, skipping", filename)
                    continue

                # Only execute if we haven't touched the 'run' file
                env = os.environ.copy()               
                env.update({
                    str('CYCLECLOUD_PROJECT_PATH'): str(os.path.join(root, self.project_name)),
                    str('CYCLECLOUD_PROJECT_NAME'): str(self.project_name),
                    str('CYCLECLOUD_SPEC_PATH'): str(os.path.join(root, self.project_name, self.name)), 
                    str('CYCLECLOUD_SPEC_NAME'): str(self.name)
                })

                # remove the internal paths to jetpack from the supplied PATH for client scripts
                if "PATH" in env:
                    env_path = env["PATH"]
                    env_path_parts = env_path.split(os.path.pathsep)
                    env_path_parts = [p for p in env_path_parts if not p.startswith("/opt/cycle/jetpack/system/") and "cycle\\jetpack\\system" not in p]
                    env_path = os.path.pathsep.join(env_path_parts)
                    env["PATH"] = env_path

                # Make the file executable
                _make_executable(os.path.join(scripts_dir, filename))

                logger.info("Executing cluster-init script: %s, output written to %s", script, output_file)
    
                p = self._run_script(script, env, scripts_dir, output_file)
      
                if p.returncode == 0:
                    logger.info("Cluster-init script %s ran successfully!", script)
                else:
                    description = "Failed to execute cluster-init script '%s' in project '%s' (return code: %s)" % (script, self.project_name, p.returncode)
                    logger.error(description)
                    if fail_on_error:
                        # NOTE: We are just reading the file written to disk from "run_script", in the future we can re-write run script
                        # to use subprocess.check_output vs Popen. We didn't do this right now because it's a stable release, in 8.x or later
                        # we could/should rewrite it.
                        output = ""
                        with open(output_file) as f:
                            output = f.read()
                        detail = "Script output:\n %s" % output
                        recommendation = "Edit and re-upload the script to correct the error and try again"
                        raise jetpack.util.UserError("Error running cluster-init", description=description, detail=detail, recommendation=recommendation)
                    
                # Touch a file in the run_dir so we dont' run the script again
                open(run_file, 'w').close()


    def _run_script(self, filename, env, cwd, output_file):
        """ Runs a script and streams it's output to a file, returns the process """
        with open(output_file, 'w') as f:
            
            try:
                # NOTE: we run different Popen commands on windows vs linux due to how the shells are opened
                if jetpack.util.is_windows():
                    p = Popen(filename, stdout=f, stderr=STDOUT, env=env, cwd=cwd, shell=True)
                else:
                    p = Popen([filename], stdout=f, stderr=STDOUT, env=env, cwd=cwd)                   
                p.wait()
            except OSError as e:
                message = "Error running cluster-init"
                description = str(e) + " while executing %s in project %s" % (os.path.basename(filename), self.project_name)
                if e.errno in [2, 8] and not jetpack.util.is_windows():
                    recommendation = "Check that the shebang line (#!) exists and points to a valid interpreter"
                    # Check for windows style line endings (\r\n vs \n)
                    if self._has_windows_line_endings(filename):
                        recommendation = "Check that the script doesn't contain Windows line endings"
                    raise jetpack.util.UserError(message, description=description, recommendation=recommendation)
                else: 
                    raise
        return p

    def _has_windows_line_endings(self, filename):
        """ Check if a file contains \r\n meaning it has windows style line endings """
        with open(filename) as f:
            return '\r\n' in f.readline()
        return False

    def _create_directories(self, root):
        """ Create the directories required to run the cluster init """
        run_dir = os.path.join(root, '.run', self.project_name, self.name, 'scripts')
        scripts_dir = os.path.join(root, self.project_name, self.name, 'scripts')
        log_dir = os.path.join(jetpack.util.jetpack_home(), 'logs', 'cluster-init', self.project_name, self.name, 'scripts')

        # Ensure the run directory exists
        if not os.path.isdir(run_dir):
            os.makedirs(run_dir)
        if not os.path.isdir(log_dir):
            os.makedirs(log_dir)
        if not os.path.isdir(scripts_dir):
            os.makedirs(scripts_dir)

        return (run_dir, scripts_dir, log_dir)


def _get_cluster_init_home():
    '''This is a helper for the unit tests'''
    return jetpack.config.get('cluster_init.home_dir')


def get_specs(config=None, project_name: str = None, name: str = None) -> List[ClusterInitSpec]:
    """ Returns a list of ClusterInitSpec objects based on the userdata """

    # Read the config and see if cyclecloud.project_specs exists
    specs = jetpack_config.get('cyclecloud.cluster_init_specs', [])

    # Get all of the specs except the one named cyclecloud, this one is no longer needed. It shouldn't be passed in
    # but there could be a case where an older CycleCloud is starting a newer jetpack
    specs = [ClusterInitSpec(spec, config=config, load_from_manifest=True) for spec in specs if spec.get('project', '').lower() != "cyclecloud"]
    if project_name:
        specs = [spec for spec in specs if spec.project_name == project_name]
        if name:
            specs = [spec for spec in specs if spec.name == name]

    return specs


def _process_chef_tarball(src_dir, tarball, dst_dir, folder): 
    try:
        tmp_dir = tempfile.mkdtemp()

        if not os.path.isfile(os.path.join(src_dir, tarball)):
            return
        with tarfile.open(os.path.join(src_dir, tarball), 'r:gz') as tgz:
            tgz.extractall(tmp_dir)
        
        subdir, _ = os.path.splitext(tarball)

        # Ensure the destination exists
        if not os.path.isdir(os.path.join(dst_dir, subdir)):
            os.makedirs(os.path.join(dst_dir, subdir))

        for filename in os.listdir(os.path.join(tmp_dir, subdir)):
            src = os.path.join(tmp_dir, subdir, filename)
            dst = os.path.join(dst_dir, folder, filename)
            _move_and_overwrite(src, dst)
    finally:
        shutil.rmtree(tmp_dir)

def _move_and_overwrite(src, dst):
    """ Copes a file or directory from src to dst, if it's a directory and it already exists
    remove it first """
    if os.path.exists(dst):
        if os.path.isdir(dst):
            shutil.rmtree(dst)

    logger.debug("Copying %s => %s", src, dst)
    shutil.move(src, dst)
    
def _make_executable(path):
    """ Given a path to a file, make that file executable """
    st = os.stat(path)
    os.chmod(path, st.st_mode | stat.S_IEXEC)

def process_scripts(no_sync=False, rerun=False, config=None):
    """ Runs the clusterinit(s) in the appropriate order
    
    When no_sync is true, the files are not sync'd from cloud storage
    When rerun is true, all executables will be re-run regardless if they have already completed successfully
    """

    try:
        logger.info("Executing cluster-init...")
        specs = get_specs(config)

        if no_sync is False:
            for spec in specs:
                spec.download_scripts()

        for spec in specs:
            spec.run(rerun=rerun)
        logger.info("Completed cluster-init!")
        jetpack.status.update(jetpack.status.JetpackResource.SoftwareConfiguration, jetpack.status.JetpackStatus.Ok, "Cluster-init completed successfully")
    except jetpack.util.UserError as e:
        jetpack.status.update_from_exception(jetpack.status.JetpackResource.SoftwareConfiguration, e)
        raise ClusterInitError(str(e))
    except Exception as e:
        jetpack.status.update_from_exception(jetpack.status.JetpackResource.SoftwareConfiguration, e)
        raise
    finally:
        jetpack.status.send()

def process_chef(config=None):
    """ Iterates over all specs assigned to the node:
      - download
      - extract
      - move to $JETPACK_HOME/system/chef/chef-repo/cookbooks|roles|data_bags
        ** In the case of a directory, the existing directory will be removed before the copy
    """

    specs = get_specs(config)
    for spec in specs:
        spec.download_chef()

def process_manifests(config=None):
    pass
    # TODO: Get all the manifests!?

def download_blob(path, dst, project, spec=None):
    """ Given a project name (and possibly spec name), download a blob
    specified by 'path' (relative path) to dst

    Example: download_blob("files/myfile.tgz", "~/Downloads", project="cyclecloud")
    
    Assuming project=cyclecloud had a base location of az://account/container/data, the blob would be
    downloaded from az://account/containers/data/files/myfile.tgz to ~/Downloads"""
    specs = get_specs(project_name=project, name=spec)
    if not specs:
        raise ClusterInitError("Unknown project '%s'!" % project)
    
    spec = specs[0]

    project_name = urllib.parse.quote(project, safe="")
    spec_name = urllib.parse.quote(spec.name, safe="")
    project_version = urllib.parse.quote(spec.project_version, safe="")

    # self.location is the path to the spec "<base>/projects/<name>/<version>/<spec>" but we want the path to the actual project files
    # so we have to remove the <version>/<spec> and add on files
    project_blob_src = spec.location[:-len("/%s/%s" % (project_version, spec_name))]
    project_blob_src = "%s/blobs/%s" % (project_blob_src, path)
    
    user_blob_src = "blobs/%s/%s" % (project_name, path)

    # Attempt to download the user blob first, if it doesn't exist then try to download the project blob
    try:
        spec._download(user_blob_src, dst, verify_source=True)
        logger.info("Downloaded %s" % user_blob_src)
    except SourceNotFound:
        logger.info("Could not find user blob '%s', checking project blobs" % user_blob_src)
        spec._download(project_blob_src, dst, verify_source=True)
        logger.info("Downloaded %s" % project_blob_src)    


class JetpackTestException(Exception):
    pass

        
def send_test_report(report_msg):
    jetpack.messaging.send_internal_message(report_msg, "test")


def _get_cluster_name():
    return jetpack.config.get('identity.cluster_name')


def _get_instance_id():
    return jetpack.config.get('identity.instance_id')


class JetpackTestReport:

    def __init__(self):
        self._exitstatuses = []
        self._report = {
            'created_at': None,
            'status': 'notests',
            'summary': {'duration': 0.0, 'failed': 0, 'num_tests': 0, 'error': 0,
                        'passed': 0, 'skipped': 0},
            'tests': []}

    @staticmethod
    def from_json(json_string):
        _report_dict = json.loads(json_string)
        report = JetpackTestReport()
        report._report = _report_dict
        return report
        
    @property
    def num_tests(self):
        return self._report['summary']['num_tests']

    @property
    def num_skipped(self):
        return self._report['summary']['skipped']

    @property
    def failed(self):
        return self._report['summary']['failed'] > 0 or self._report['summary']['error'] > 0
        
    @property
    def num_failed(self):
        return self._report['summary']['failed']

    def error_message(self):
        if self.num_failed > 0:
            return "%d test(s) failed out of %d" % (self.num_failed, self.num_tests)
        
        if self._report['summary']['error'] > 0:
            return "%d test(s) errored out of %d" % (self._report['summary']['error'], self.num_tests)

        raise RuntimeError("No tests have failed. This method has been invoked incorrectly")

    def add(self, exitstatus, report):
        '''Update the composite test report with a report for an individual spec'''
        self._exitstatuses.append(exitstatus)

        # only update the created_at field once
        if self._report.get("created_at") is None:
            self._report["created_at"] = report["created_at"]

        current_summary = self._report['summary']
        for key in ['duration', 'failed', 'num_tests', 'passed', 'skipped', 'error']:
            current_summary[key] = current_summary[key] + report['summary'].get(key, 0)

        # update the report status    
        if current_summary['failed'] > 0:
            self._report['status'] = 'failed'
        elif current_summary['error'] > 0 and self._report['status'] != 'failed':
            self._report['status'] = 'error'
        elif current_summary['skipped'] != 0 and current_summary['skipped'] == current_summary['num_tests']:
            self._report['status'] = 'skipped'
        elif report['summary']['num_tests'] > 0:
            self._report['status'] = 'passed'

        self._report['tests'].extend(report.get('tests', []))

    def to_message(self):
        '''Convert the composite report to the 'data' portion of a system message
        so that it conforms to the system message schema
        https://internal-docs.cyclecomputing.com/cyclecloud/latest/internalDevHandbook/system_messages.html#systemmessage-schema
        '''
        return self._report

    def to_json(self):
        '''Convert the report to json document'''
        return json.dumps(self._report)


def _get_results_filepath(logdir=None):
    if logdir is None:
        logdir = os.path.join(jetpack.util.jetpack_home(), 'logs')
    return os.path.join(logdir, 'test-results.json')


def run_tests(capture=True, send_report=True):
    '''
    Run integration tests for cluster-init

    arguments:
      capture -- capture stdout and stderr during test run rather than printing to console
      send_report -- whether or not to send results back to CycleServer
    '''
    report = JetpackTestReport()

    specs = get_specs()
    for s in specs:
        status, spec_report = s.test(capture=capture)
        report.add(status, spec_report['report'])
    
    if report.num_tests == 0:
        logger.warning("No tests found")
    elif report.num_tests == report.num_skipped:
        logger.warning("Tests found but all skipped")

    with open(_get_results_filepath(), 'w') as f:
        f.write(report.to_json())

    if send_report:
        send_test_report(report.to_message())

    return report


def repost_test_report():
    '''
    Repost the results of a previous test run
    The test results are reposted to avoid a race condition where
    the initial test run succeeds but a subsequent chef converge succeeds immediately after
    Instead of posting a test failure and then a converge success, we post two test failure
    '''
    report_path = _get_results_filepath()
    with open(report_path, 'r') as f:
        report = JetpackTestReport.from_json(f.read())
            
    send_test_report(report.to_message())

    return report
