import datetime
import os
import re
import subprocess
import tempfile
from xml.etree import ElementTree as ET


class SpecNotFound(Exception):
    pass


def report_from_xml(f, project, spec):
    """ Generate a report datastructure that CycleCloud expects from a junit XML file

    NOTE: The report structure CC expects is actually the pytest internal representation of a testsuite, we convert
    due to legacy behavior of how the automated testing works (i.e. we used to call pytest internal methods to run the testsuite vs
    just shelling out and getting xml)"""
    tree = ET.parse(f)
    root = tree.getroot()

    # NOTE: Depending on the version of pytest being used, the junit xml will 
    # start with either "testsuites" containing a list of testsuites
    # or a single "testsuite"
    if root.tag == "testsuites":
        if len(root) != 1:
            raise Exception("Didn't find exactly one testsuite, found %s" % len(root))
        testsuite = root[0]
    elif root.tag == "testsuite":
        testsuite = root
    else:
        raise Exception("Unexpected root tag in junit xml: %s" % root.tag)

    passed = 0
    failed = int(testsuite.attrib['failures'])
    error = int(testsuite.attrib['errors'])

    # junit xml changed between pytest 3.0.4 and newer
    if 'skips' in testsuite.attrib:
        skipped = int(testsuite.attrib['skips'])
    elif 'skipped' in testsuite.attrib:
        skipped = int(testsuite.attrib['skipped'])

    tests = {}

    for case in testsuite:
        # a testcase with no children is a success
        # otherwise we have to read the first child for more info

        # NOTE: We aren't filling out the setup, call and teardown methods exactly as pytest would
        # because we only use the longrepr in CC
        name = "%s::%s" % (case.attrib['classname'], case.attrib['name'])

        if name not in tests:
            tests[name] = {
                "name": name,
                "run_index": 0,  # hard coded?
                "project": project,
                "spec": spec,
                "outcome": None,  # failed, passed, skipped, error
                "duration": float(case.attrib['time']),
                "setup": {},
                "call": {},
                "teardown": {}
            }

        test = tests[name]

        if len(case) > 0:
            reason = case[0]
            if reason.tag == "failure":
                test["outcome"] = 'failed'
                test["call"] = {
                    'outcome': 'failed',
                    'longrepr': reason.attrib['message']
                }
            elif reason.tag == "skipped":
                test["outcome"] = 'skipped'
                test["setup"] = {
                    'outcome': 'skipped',
                    'longrepr': reason.attrib['message']
                }
            elif reason.tag == "error":
                test["outcome"] = "failed" if test["outcome"] == "failed" else "error"
                setup_heuristic = re.search(r"def\s+setup\s*\(", reason.text, re.MULTILINE)
                teardown_heuristic = re.search(r"def\s+teardown\s*\(", reason.text, re.MULTILINE)
                if teardown_heuristic and not setup_heuristic:
                    test["teardown"] = {
                        'outcome': 'error',
                        'longrepr': reason.text
                    }
                else:
                    test["setup"] = {
                        'outcome': 'error',
                        'longrepr': reason.text
                    }
            else:
                raise Exception("Unknown test result: '%s'" % reason.tag)
        else:
            test["outcome"] = 'passed'
            passed += 1

    report = {
        "report": {
            "created_at": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f"),
            "summary": {
                "passed": passed,
                "failed": failed,
                "error":  error,
                "skipped": skipped,
                "num_tests": len(tests), # number of tests is reported incorrectly when an error occurs
                "duration": float(testsuite.attrib['time'])
            },
            "tests": list(tests.values())
        }
    }
    return report


def run(specdir, project, spec, capture=True):
    """ Shell out to pytest and parse the junit XML into a report structure expected by CycleCloud """
    # Shell out to pytest and write xml to tempfile

    if not os.path.exists(specdir):
        raise SpecNotFound("No Spec exists at path %s" % specdir)

    _, xml_output = tempfile.mkstemp()
    try:
        cmd = ['pytest', specdir, '--junit-xml=%s' % xml_output]
        if capture is False:
            cmd.append("-s")

        exitstatus = subprocess.call(cmd)
        if os.path.isfile(xml_output):
            report = report_from_xml(xml_output, project, spec)
        else:
            raise Exception("pytest failed to run tests: %s" % exitstatus)
        return exitstatus, report
    finally:
        if os.path.isfile(xml_output):
            os.remove(xml_output)
