# This pytest plugin reports project spec tests as a Python dictionary
# This plugin annotates discovered tests w/ project and spec names
#

import datetime
import platform
import re
import time


class SpecReport(object):
    def __init__(self, project, spec):
        self.project = project
        self.spec = spec
        self.nodes = {}
        self.summary = {}
        self.run_index = 0
        self.report = {}

    def _get_outcome(self, report):
        if report.failed:
            if report.when != 'call':
                return 'error'
            else:
                if hasattr(report, 'wasxfail'):
                    return 'xpassed'
                else:
                    return 'failed'
        elif report.skipped:
            if hasattr(report, 'wasxfail'):
                return 'xfailed'
            else:
                return 'skipped'

        return report.outcome

    def _update_summary(self, outcome, report):
        if report.passed and report.when != 'call':
            # do not count successful setup and teardown
            return
        
        if outcome not in self.summary:
            self.summary[outcome] = 1
        else:
            self.summary[outcome] += 1

    def _make_stage_dict(self, report):
        outcome = self._get_outcome(report)
        
        stage_dict = {
            'name': report.when,
            'duration': getattr(report, 'duration', 0.0),
            'outcome': outcome
        }

        if hasattr(report, 'stage_metadata'):
            stage_dict['metadata'] = report.stage_metadata

        if hasattr(report, 'wasxfail'):
            stage_dict['xfail_reason'] = report.wasxfail

        if report.longrepr:
            longrepr = report.longrepr
            # the full longrepr prints a large stack traceback
            # that drowns the actual error output
            if hasattr(report.longrepr, 'reprcrash'):
                longrepr = report.longrepr.reprcrash.message

            stage_dict['longrepr'] = str(longrepr)

        # only show stdout/stderr for the current stage
        stage_matcher = re.compile(r'^Captured (.+) {}$'.format(report.when))
        for header, content in report.sections:
            match = stage_matcher.match(header)
            if match:
                stage_dict[match.group(1)] = content

        return stage_dict

    def pytest_runtest_logreport(self, report):
        stage_dict = self._make_stage_dict(report)

        self._update_summary(stage_dict['outcome'], report)

        if hasattr(report, 'report_metadata'):
            if 'metadata' not in self.report:
                self.report['metadata'] = []

            self.report['metadata'].append(report.report_metadata)

        if report.nodeid not in self.nodes:
            self.nodes[report.nodeid] = {
                'name': report.nodeid,
                'duration': stage_dict['duration'],
                'run_index': self.run_index
            }
            self.run_index += 1

        self.nodes[report.nodeid][report.when] = stage_dict
        self.nodes[report.nodeid]['duration'] += stage_dict['duration']

        # do this after we know we've created the nodeid in reports
        if hasattr(report, 'test_metadata'):
            if 'metadata' not in self.nodes[report.nodeid]:
                self.nodes[report.nodeid]['metadata'] = []

            self.nodes[report.nodeid]['metadata'].append(
                report.test_metadata)

    def pytest_sessionstart(self, session):
        self.session_start_time = time.time()

    def _get_overall_outcome(self, report):
        if report['setup']['outcome'] != 'passed':
            return report['setup']['outcome']

        # if the call has passed, just return the outcome of teardown (which
        # might be an error)
        if report['call']['outcome'] == 'passed':
            return report['teardown']['outcome']

        return report['call']['outcome']

    def _default_report(self, env, tests, created_at):
        report = {
            'report': {
                'environment': env,
                'tests': tests,
                'summary': self.summary,
                'created_at': created_at
            }
        }

        if 'metadata' in self.report:
            report['report']['metadata'] = self.report['metadata']

        return report

    def pytest_sessionfinish(self, session):
        session_stop_time = time.time()
        session_duration = session_stop_time - self.session_start_time

        env = {}
        env['Python'] = platform.python_version()
        env['Platform'] = platform.platform()

        created = datetime.datetime.now()
        created_at = str(created)

        self.summary['num_tests'] = len(self.nodes)
        self.summary['duration'] = session_duration

        # format the reports
        tests = []
        for test, report in list(self.nodes.items()):
            report['outcome'] = self._get_overall_outcome(report)
            report['project'] = self.project
            report['spec'] = self.spec
            tests.append(report)

        report = self._default_report(env, tests, created_at)

        return report
