from packaging import version
import distro  # type: ignore
import json
import logging
import os
import re
import shutil
import subprocess
from typing import Optional
import urllib.request
import urllib.parse
import urllib.error


import jetpack.util
import jetpack.util.system
import jetpack.settings
import jetpack.config
import jetpack.status

from jetpack.status import JetpackResource, JetpackStatus

if jetpack.util.is_linux():
    import grp
    import pwd

logger = logging.getLogger(__name__)

class AddUserException(Exception):
    pass

class UserAlreadyExists(Exception):
    pass

class AddGroupException(Exception):
    pass

class GroupAlreadyExists(Exception):
    pass

class UserManagementException(Exception):
    pass

class MountedFilesystemUnavailable(Exception):
    pass

class InvalidConfiguration(Exception):
    """ Raised when an invalid configuration is encountered that cannot be rectified """
    pass

# This is a blacklist of usersnames we do not allow to be managed
USERNAME_BLACKLIST = ['root', 'admin']

DEFAULT_SHELL = '/bin/bash'
NOLOGIN_SHELL = '/sbin/nologin'

LOGIN_DEFS_PATH = '/etc/login.defs'
ADDUSER_CONF_PATH = '/etc/adduser.conf'
UID_MAX = 19999
GID_MAX = 19999


def query(is_admin=None, is_owner=None):
    """ Public API: Reads the list of users from disk and returns them, in the future we may add filters here """
    auth = _read_local_auth()
    users = auth.get('users', [])

    if is_admin is not None:
        users = [user for user in users if user.get('isAdmin', False) == is_admin]
    if is_owner is not None:
        users = [user for user in users if user.get('isOwner', False) == is_owner]

    return users
        
def owner():
    """ Public API: Returns the cluster owner
    
    Raises exception if more than one owner failed, None if no owners otherwise a singular dictionary representing the owner"""
    owners = query(is_owner=True)
    if len(owners) > 1:
        raise UserManagementException("More than one (%s) owner found" % len(owners))
    elif len(owners) == 1:
        return owners[0]
    else:
        return None

@jetpack.util.run_once
def _update_max_uid_gid():
    """ Sets UID_MAX/GID_MAX in /etc/login.defs and LAST_UID/LAST_GID in 
    /etc/adduser.conf (for debian-based systems) to 19999 so our managed
    users can use UIDs 20000 and above without causing conflicts.

    Note that if the files don't exist, they will not be created.
    
    Returns a tuple of 2 booleans, one that indicates whether /etc/login.defs
    was updated, and one that indicates whether /etc/adduser.conf was updated.
    """
    return (_update_logindefs(), _update_adduserconf())

def _update_logindefs():
    if os.path.exists(LOGIN_DEFS_PATH):
        return LoginDefsGenerator(LOGIN_DEFS_PATH, UID_MAX, GID_MAX).update()
    return False

def _update_adduserconf():
    if os.path.exists(ADDUSER_CONF_PATH):
        return AddUserConfGenerator(ADDUSER_CONF_PATH, UID_MAX, GID_MAX).update()
    return False

@jetpack.util.run_once
def _configure_home_directory():
    """ Configures the home directory: makes sure the filesystem is mounted, then creates the homedir  and validates any selinux
    policies that might to be set """
    home = jetpack.config.get('cuser.base_home_dir', '/home')

    # Check to see if the home directory is a mounted filesystem and if it is... is it to be ready?
    if not _mountpoint_ready(home):
        raise MountedFilesystemUnavailable("The home directory mounted at %s is not available" % home)

    # Create any subdirs as needed and make sure they are correctly permissioned    
    if not os.path.isdir(home):
        try:
            os.makedirs(home)
        except OSError as e:
            if e.errno == 17:  # Err 17 is that the file already exists, so we're good if some other box already created it
                pass
            else:
                raise
    jetpack.util.update_file_permissions(home, mode=0o755)

    # Default cyclecloud policy should be enforcing
    if jetpack.util.get_selinux_enforce() and jetpack.util.system.platform_family() == 'rhel':
        # Cannot set file context equivalency to self, if the directory is /home
        if home != "/home":
            # Follow symlink with realpath
            realhome = os.path.realpath(home)

            # Ensure we have the proper commands available to allow for a non-standard NFS home directory
            jetpack.util.system.install_package("policycoreutils") # provides `restorecon` and `setsebool`

            if version.parse(distro.version()) < version.parse("8.0"):
                jetpack.util.system.install_package("policycoreutils-python")  # provides `semanage`
            else: 
                jetpack.util.system.install_package("policycoreutils-python-utils") # provides `semanage`

            # Check if realhome already has file context equivalency to /home
            if subprocess.call("semanage fcontext -l -C | grep -w '%s = /home'" % realhome, shell=True) != 0:
                # Set /home equivalency to realhome
                logger.debug("Calling 'semanage fcontext -a -e /home %s'" % realhome)
                subprocess.check_call("semanage fcontext -a -e /home %s" % realhome, shell=True)

                # Restore file context recursive to realhome
                logger.debug("Calling 'restorecon -R %s" % realhome)
                subprocess.check_call("restorecon -R %s" % realhome, shell=True)
        
        # If the home directory is a nfs mount
        mount = _cyclecloud_mount(home)
        if mount is not None and mount.get('type') == 'nfs':
            logger.debug("Calling 'setsebool -P use_nfs_home_dirs 1'")
            subprocess.check_call('setsebool -P use_nfs_home_dirs 1', shell=True)

def manage():
    """ Public API: Runs the user management operation """
    if jetpack.settings.get('authenticator.method', None) != "internal":
        logger.debug("authenticator.method is not 'internal' - no need to manage users via jetpackd")
        return

    if jetpack.util.is_windows():
        return

    # Configure the home directory we'll be putting our users into
    try:
        _configure_home_directory()
    except MountedFilesystemUnavailable as e:
        # If the home directory isn't mounted yet this is an error
        logger.error(str(e))
        return [str(e)]

    # Set max UID/GID so we can use values above that for our users
    uid_gid_updated = _update_max_uid_gid()
    if uid_gid_updated:
        if uid_gid_updated[0]:
            logger.debug("Updated UID/GID max values in /etc/login.defs")
        if uid_gid_updated[1]:
            logger.debug("Updated UID/GID max values in /etc/adduser.conf")

    # Track error messages so we can send the appropriate status back to CycleCloud at the end of user management
    errors = []

    rerun = False
    managed_auth = _read_managed_auth()
    etag = managed_auth.get('etag', None)

    result = _read_remote_auth(etag=etag)

    # If the auth file didn't change and we don't want to re-run, just exit early cause our work here is done
    if result is None and not managed_auth.get('rerun', False):
        logger.debug("Remote auth.json is unchanged, continuing...")
        return

    # Store the returned etag so we can query only for changes next iteration
    if result is not None:
        managed_auth['etag'] = result.get('etag', None)
        auth = result.get('auth', {})
    else:
        # If we didn't get a result back it means we didn't pull a file from storage but we are still running due to a rerun request
        # So read the local file instead
        auth = _read_local_auth()

    managed_users_by_username = {}
    users_by_username = {}
    target_version = auth.get('version')

    if target_version:
        logger.info("Applying version %d of user list" % target_version)
    else:
        logger.info("Applying unversioned user list")

    for user in managed_auth.get('users', []):
        managed_users_by_username[user['name']] = user
    for user in auth.get('users', []):
        users_by_username[user['name']] = user
    
    for username in users_by_username.keys():
        if username in USERNAME_BLACKLIST:
            message = "Username '%s' is not allowed" % username
            errors.append(message)
            logger.error(message)
            continue
        
        # Keep a running collection of keys we need to add at the end
        authorized_keys = {}

        # Add any users that aren't already managed
        if username not in managed_users_by_username:
            user = users_by_username[username]

            # TODO: Add proper group support instead of matching the 
            #       user's name and uid
            groupname = username
            gid = user['uid']

            try:
                _add_group(groupname, gid)
                logger.info("Added group `%s`", groupname)
            except GroupAlreadyExists:
                logger.info("Group `%s` already exists on the system, not adding", groupname)
            except AddGroupException as e:
                message = "Unable to add group '%s': %s" % (groupname, str(e))
                logger.error(message)
                errors.append(message)
                continue
            except Exception as e:
                message = "Error adding group '%s': %s" % (groupname, str(e))
                logger.exception(message)
                errors.append(message)
                continue

            try:
                _add_user(users_by_username[username], gid)
                logger.info("Added user `%s`", username)
            except UserAlreadyExists:
                logger.info("User `%s` already exists on the system, not adding", username)
            except AddUserException as e:
                message = "Unable to add user '%s': %s" % (username, str(e))
                logger.error(message)
                errors.append(message)
                continue
            except Exception as e:
                message = "Error adding user '%s': %s" % (username, str(e))
                logger.exception(message)
                errors.append(message)
                continue

            # First time we add a user, generate a public key for them and add it to the authorized keys file
            home, uid, gid, _ = _user_information(username)
            ssh_dir = os.path.join(home, '.ssh')
            try:
                _ensure_directory(ssh_dir, uid, gid, 0o700)
            except InvalidConfiguration as e:
                # Proceed to the next user since we can't continue with this user since there was a configuration error (most likely a UID/GID mismatch)
                logger.error(str(e))
                errors.append(str(e))
                continue  
    
            private_key_path = os.path.join(ssh_dir, 'id_rsa')
            keypair = SSHKeypairGenerator(private_key_path, uid, gid)
            keypair.update()

            # NOTE: We use the '' identifier because these keys aren't group managed, just simply appended to the file if they do not exist
            authorized_keys[''] = [keypair.public_key]
            
            # We also have to write/update the ~/.ssh/ssh_config file to disable strict host checking (in a threadsafe way).
            SSHConfigGenerator(os.path.join(home, '.ssh', 'config'), uid, gid).update()


        # Ensure we have a record within managed users
        if username not in managed_users_by_username:
            managed_users_by_username[username] = {}

        # If the set of keys is different than last time, update the public keys
        if set(managed_users_by_username[username].get('publicKeys', [])) != set(users_by_username[username].get('publicKeys', [])):
            # NOTE: The identifier we use to group the keys passed in is `cyclecloud:username` 
            authorized_keys["cyclecloud:%s" % username] = users_by_username[username].get('publicKeys', [])

        # Renable a previously disabled user by changing their shell back to what it was before and flagging them as not disabled
        if managed_users_by_username[username].get('disabled', False):
            if _update_shell(username, managed_users_by_username[username].get('shell', DEFAULT_SHELL)):
                managed_users_by_username[username]['disabled'] = False
                logger.info("Activated login for user %s" % username)
        
        # If their sudo access has changed since last time, update their access
        if users_by_username[username].get('isAdmin', False) != managed_users_by_username[username].get('isAdmin', False):
            is_admin = users_by_username[username].get('isAdmin', False)
            _update_sudo_access(username, sudoer=is_admin)
            
            # If isAdmin isn't defined, undefine it in managed_users_by_username
            if 'isAdmin' not in users_by_username[username]:
                managed_users_by_username[username].pop('isAdmin', None)

        # If there are any keys, update them in bulk for the user 
        if authorized_keys:
            if _update_public_keys(username, authorized_keys):
                logger.info("Updated public keys for user %s" % username)
                
        # Update our managed users record for the user with the newly passed in data
        managed_users_by_username[username].update(users_by_username[username])
        

    # To disable users we have to see users who we previously were managing who are not in the list we currently are processing
    # We can do this by a set different between the list of managed users who aren't disabled and the list of users passed in
    disabled_usernames = set(managed_users_by_username.keys()) - set(users_by_username.keys())
    for username in disabled_usernames:
        # If the user isn't currently disabled, disable them
        if managed_users_by_username[username].get('disabled', False) is False:
            _, _, _, shell = _user_information(username)
            managed_users_by_username[username]['shell'] = shell
            if _update_shell(username, NOLOGIN_SHELL):
                managed_users_by_username[username]['disabled'] = True
                logger.info("Disabled login for user %s" % username)

    # If we need to re-run to handle any users that are waiting on something (i.e. shared drives), set the flag to true so we know
    # to re-process the auth file even though it may not have changed
    if rerun:
        managed_auth['rerun'] = True
    else:
        # Otherwise we don't even need to save it
        if 'rerun' in managed_auth:
            del managed_auth['rerun']
    managed_auth['users'] = list(managed_users_by_username.values())
    _write_managed_auth(managed_auth)

    # Update the status of user management for this node on CycleCloud
    state = JetpackStatus.Ok
    message = "Updated users"
    description = None
    detail = None
    if errors:
        state = JetpackStatus.Error
        message = "Error updating users"
        
        # If there is only one error, put it in the description
        # If there are more than one error, list all the errors in the detail and have the description just
        # say how many errors there are.
        if len(errors) == 1:
            description = errors[0]
        else:
            description = "%s errors occurred while updating users" % len(errors)
            detail = "\n".join(errors)

    jetpack.status.update(JetpackResource.UserAccess, status=state,
                          message=message,
                          description=description, 
                          detail=detail,
                          version=target_version)

    return errors


def _read_local_auth():
    """ Reads the local auth file from disk """
    auth = {}
    path = os.path.join(jetpack.util.jetpack_home(), 'config', 'auth.json')
    if os.path.isfile(path):
        try:
            with open(path) as f:
                auth = json.loads(f.read())
        except ValueError as e:
            logger.warning("Unable to read auth.json - %s" % str(e))
    return auth


def _read_remote_auth(etag: str = None) -> Optional[dict]:
    """ Fetches the remote auth to the local disk and returns the etag, or None of the file hasn't been changed """
    """ Gets the remote auth file to the local disk and returns the contents if it has been updated along with etag, {} if no changes """
    auth_method = jetpack.settings.get('authenticator.method', None)
    local_auth_path = os.path.join(jetpack.util.jetpack_home(), 'config', 'auth.json')

    if auth_method == "internal":
        configuration = jetpack.util.parse_config(None)
        cluster_name = configuration['identity']['cluster_name']
        instance_id = configuration['identity']['instance_id']

        url = "/clusterlink/auth/%s" % urllib.parse.quote(instance_id)

        params = {}
        params['cluster'] = cluster_name

        # If we have previously downloaded the file, comapre the etag to see if we need to download it again
        # If we haven't downloaded the file (or it was deleted for some reason, always re-download it)
        headers = {}
        if os.path.isfile(local_auth_path):
            if etag:
                headers["If-None-Match"] = '"%s"' % etag

        logger.info("Getting latest auth configuration from %s" % url)
        r = jetpack.util.query_cyclecloud(url, params=params, headers=headers)

        if r.status == 304:
            return None
        elif r.status == 401:
            logger.error("An authentication error (%s) occurred while fetching the user list from CycleCloud: %s" % (r.status, r.reason))
            return None
        elif r.status != 200:
            logger.error("An error occurred getting auth %s: %s (%s)" % (url, r.reason, r.status))
            raise Exception(r.reason)

        data = r.read().decode()

        result = {}
        etag = r.getheader("ETag")
        if etag:
            result['etag'] = etag.strip('"')

        try:
            result['auth'] = json.loads(data)
        except ValueError:
            logger.error("auth.json was not correctly formatted JSON")
            return None

        with os.fdopen(os.open(local_auth_path, os.O_WRONLY | os.O_CREAT, 0o640), 'w') as f:
            f.write(data)
        return result
    return None


def _read_managed_auth():
    managed_auth = {}
    managed_auth_path = os.path.join(jetpack.util.jetpack_home(), 'config', 'auth.managed.json')
    if os.path.isfile(managed_auth_path):
        with open(managed_auth_path) as f:
            managed_auth = json.loads(f.read())
    return managed_auth


def _write_managed_auth(managed_auth):
    # Write the managed auth file to disk. First write to a temp file then move it to the real location so we can
    # avoid disk errors while writing
    managed_auth_path = os.path.join(jetpack.util.jetpack_home(), 'config', 'auth.managed.json')
    tmp_managed_auth_path = managed_auth_path + ".tmp"
    with os.fdopen(os.open(tmp_managed_auth_path, os.O_WRONLY | os.O_CREAT, 0o640), 'w') as f:
        f.write(json.dumps(managed_auth, indent=4, sort_keys=True))
    shutil.move(tmp_managed_auth_path, managed_auth_path)


def _update_shell(username, shell):
    """ Update the user's shell """
    try:
        subprocess.check_output(['usermod', '--shell', shell, username], stderr=subprocess.STDOUT)
        return True
    except subprocess.CalledProcessError as e:
        if e.output is not None:
            e.output = e.output.strip()
        logger.error("Failure to run `usermod`: %s (%s)" % (e.output, e.returncode))
        return False


def _add_user_to_group(username, group):
    """ Adds USERNAME to GROUP if they aren't already a member """
    members = []
    try:
        members = grp.getgrnam(group).gr_mem
    except KeyError:
        raise UserManagementException("Unable to add `%s` to group `%s` because the group does not exist, please check your Jetpack installation." % (username, group))

    # Only add them if they aren't already in the group
    if username not in members:
        try:
            subprocess.check_output(['usermod', '-a', '-G', group, username], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            if e.output is not None:
                e.output = e.output.strip()
            logger.exception("Failure to add `%s` to group `%s`: %s (%s)" % (username, group, e.output, e.returncode))

def _remove_user_from_group(username, group):
    """ Removes USERNAME from GROUP if they are currently a member """
    members = []
    try:
        members = grp.getgrnam(group).gr_mem
    except KeyError:
        raise UserManagementException("Unable to remove `%s` from group `%s` because the group does not exist, please check your Jetpack installation." % (username, group))

    # Only remove them if they are a member
    if username in members:
        try:
            subprocess.check_output(['gpasswd', '-d', username, group], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            if e.output is not None:
                e.output = e.output.strip()
            logger.exception("Failure to remove `%s` from group `%s`: %s (%s)" % (username, group, e.output, e.returncode))

def _update_sudo_access(username, sudoer=False):
    """ Granting a removing sudo access is done via the `cyclecloud` group. If sudoer==True then the user should be part of the
    group, if sudoer=False the user should not be part of the group 
    
    Note: the cyclecloud group is installed and configured as part of jetpack setup
    """
    if sudoer:
        _add_user_to_group(username, 'cyclecloud')
        logger.info("Granted sudo access to `%s`", username)
    else:
        _remove_user_from_group(username, 'cyclecloud')
        logger.info("Revoked sudo access from `%s`", username)

def _ensure_directory(d, uid, gid, perms):
    """ Ensure that a directory exists and is owned and permissioned correctly """
    if not os.path.isdir(d):
        os.makedirs(d)
        jetpack.util.update_file_permissions(d, uid, gid, perms)
    else:
        info = os.stat(d)
        if info.st_uid != uid or info.st_gid != gid:
            raise InvalidConfiguration("The directory '%s' does not have the expected uid/gid! (expected %s/%s found %s/%s)" % (d, uid, gid, info.st_uid, info.st_gid))

def _add_user(user, gid):
    username = user['name']
    fullname = user.get('fullName', '')
    uid = None

    # Get the home directory and always resolve it back to the "real" path in case of a symlink
    # This is because selinux doesn't operate well on symlinks and will not properly flag directories like ~/.ssh otherwise
    home_dir = os.path.realpath(jetpack.config.get('cuser.base_home_dir', '/home'))

    # Verify that the user doesn't already exist on the system
    try:
        pwd.getpwnam(username)
        raise UserAlreadyExists("The user `%s` already exists" % username)
    except KeyError:
        pass

    # Verify that the UID is available
    try:
        uid = int(user['uid'])
    except KeyError:
        raise AddUserException("The UID for %s was not specified" % username)
    except ValueError:
        raise AddUserException("The UID %s for user %s is not valid" % (user['uid'], username))

    try:
        pwd.getpwuid(uid)
        raise AddUserException("The UID %s for user `%s` already exists" % (uid, username))
    except KeyError:
        pass

    try:
        subprocess.check_output(['useradd', '-b', home_dir,  '-u', '%s' % uid, '-g', '%s' % gid, '-m', '-c', fullname, username], stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        if e.output is not None:
            e.output = e.output.strip()
        raise AddUserException("Failure to run `useradd`: %s (%s)" % (e.output, e.returncode))

def _add_group(groupname, gid):
    # Verify that the group doesn't already exist on the system
    try:
        grp.getgrnam(groupname)
        raise GroupAlreadyExists("The group `%s` already exists" % groupname)
    except KeyError:
        pass

    # Verify that the GID is valid
    try:
        gid = int(gid)
    except ValueError:
        raise AddGroupException("The GID %s for group %s is not valid" % (gid, groupname))

    try:
        grp.getgrgid(gid)
        raise AddGroupException("The GID %s for group `%s` already exists" % (gid, groupname))
    except KeyError:
        pass

    try:
        subprocess.check_output(['groupadd', '-g', '%s' % gid, groupname], stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        if e.output is not None:
            e.output = e.output.strip()
        raise AddGroupException("Failure to run `groupadd`: %s (%s)" % (e.output, e.returncode))


def _update_public_keys(username, authorized_keys):
    home, uid, gid, _ = _user_information(username)
    authorized_keys_file = os.path.join(home, ".ssh", "authorized_keys")

    return AuthorizedKeysGenerator(authorized_keys_file, authorized_keys, uid, gid).update()

def _user_information(username):
    """ Given a username, get the home dir, UID and GID """
    home = os.path.expanduser("~%s" % username)
    info = pwd.getpwnam(username)
    uid = info.pw_uid
    gid = info.pw_gid
    shell = info.pw_shell
    return (home, uid, gid, shell)

# TODO: Move to a mountpoint library
def _cyclecloud_mount(path):
    """ Returns a mount dictionary if the path is within a cyclecloud mount, otherwise 'None' """
    selected = None
    longest_common_path = 0
    mounts = jetpack.config.get('cyclecloud.mounts',{})
    for name, mount in mounts.items():
        if mount.get('disabled', False):
           continue
        l = len(jetpack.util.commonpath([mount.get('mountpoint') or '', path]))
        if l > 1 and l > longest_common_path:
            selected = name
    if selected:
        return mounts[selected]

# TODO: Move to a mountpoint library
def _mountpoint_ready(path):
    """ Given a path, check to see if it's ready to be written to by making sure that it's not a shared filesystem waiting to be mounted """
    mount = _cyclecloud_mount(path)
    ready = False
    if mount:
        # Otherwise, we have to see if the filesystem is mounted before declaring it "ready"
        export_path = mount.get('export_path', None)
        mountpoint = mount.get('mountpoint', None)

        try:
            if mountpoint and os.path.ismount(mountpoint):
                ready = True
            elif mountpoint and export_path and os.path.samefile(mountpoint, export_path):
                # The mountpoint and export path point to the same file meaning we are probably the one exporting and mounting self
                ready = True
        except OSError as e:
            # If err 2 (no such file) we know the mountpoint isn't ready yet
            if e.errno == 2:
                ready = False
            else:
                raise
    else:
        # If it's not a cyclecloud mount, we can assume it's not a shared filesystem and thus it is ready
        ready = True

    return ready

class LoginDefsGenerator(jetpack.util.LockedFileManager):
    """ Updates UID_MAX and GID_MAX for /etc/login.defs in a threadsafe way 

    uid_max and gid_max are the new values to set for UID_MAX and GID_MAX in the file
    """

    _MODIFIED_MSG = "# This value has been automatically modified by Azure CycleCloud's user management"
    _UID_GID_PATTERN = re.compile(r'^\s*(UID_MAX|GID_MAX)\s*(\d+)', re.IGNORECASE)

    def __init__(self, filename, uid_max, gid_max):
        self.filename = filename
        self.uid_max = uid_max
        self.gid_max = gid_max

    def directory(self):
        return os.path.dirname(self.filename)

    @staticmethod
    def parse_line(line):
        m = LoginDefsGenerator._UID_GID_PATTERN.match(line)
        if m:
            return (m.group(1), int(m.group(2)))
        return None
    
    def check_for_update(self):
        needs_update = False
        with open(self.filename, 'r') as logindefs:
            for line in logindefs:
                l = LoginDefsGenerator.parse_line(line)
                if not l:
                    pass
                elif l[0] == "GID_MAX" and l[1] != self.gid_max:
                    needs_update = True
                elif l[0] == "UID_MAX" and l[1] != self.uid_max:
                    needs_update = True
        return needs_update

    def write(self):
        # Preserve existing UID/GID of /etc/login.defs
        st = os.stat(self.filename)

        modified = False
        tmp_filename = self.filename + ".tmp"
        with open(self.filename, 'r') as existing:
            with open(tmp_filename, 'w') as tmp:
                for line in existing:
                    l = LoginDefsGenerator.parse_line(line)
                    if not l:
                        tmp.write("%s\n" % line)
                    elif l[0] == "GID_MAX" and l[1] != self.gid_max:
                        modified = True
                        tmp.write("%s\n" % LoginDefsGenerator._MODIFIED_MSG)
                        tmp.write("GID_MAX %d \n" % self.gid_max)
                    elif l[0] == "UID_MAX" and l[1] != self.uid_max:
                        modified = True
                        tmp.write("%s\n" % LoginDefsGenerator._MODIFIED_MSG)
                        tmp.write("UID_MAX %d \n" % self.uid_max)
        if modified:
            jetpack.util.update_file_permissions(tmp_filename, st.st_uid, st.st_gid, 0o644)
            shutil.move(tmp_filename, self.filename)

class AddUserConfGenerator(jetpack.util.LockedFileManager):
    """ Updates LAST_UID and LAST_GID for /etc/adduser.conf in a threadsafe way 

    uid_max and gid_max are the new values to set for LAST_UID and LAST_GID in the file
    """

    _MODIFIED_MSG = "# This value has been automatically modified by Azure CycleCloud's user management"
    _UID_GID_PATTERN = re.compile(r'^\s*(LAST_UID|LAST_GID)\s*=\s*(\d+)', re.IGNORECASE)

    def __init__(self, filename, uid_max, gid_max):
        self.filename = filename
        self.uid_max = uid_max
        self.gid_max = gid_max

    def directory(self):
        return os.path.dirname(self.filename)

    @staticmethod
    def parse_line(line):
        m = AddUserConfGenerator._UID_GID_PATTERN.match(line)
        if m:
            return (m.group(1), int(m.group(2)))
        return None
    
    def check_for_update(self):
        needs_update = False
        with open(self.filename, 'r') as adduserconf:
            for line in adduserconf:
                l = AddUserConfGenerator.parse_line(line)
                if not l:
                    pass
                elif l[0] == "LAST_GID" and l[1] != self.gid_max:
                    needs_update = True
                elif l[0] == "LAST_UID" and l[1] != self.uid_max:
                    needs_update = True
        return needs_update

    def write(self):
        # Preserve existing UID/GID of /etc/adduser.conf
        st = os.stat(self.filename)

        modified = False
        tmp_filename = self.filename + ".tmp"
        with open(self.filename, 'r') as existing:
            with open(tmp_filename, 'w') as tmp:
                for line in existing:
                    l = AddUserConfGenerator.parse_line(line)
                    if not l:
                        tmp.write("%s\n" % line)
                    elif l[0] == "LAST_GID" and l[1] != self.gid_max:
                        modified = True
                        tmp.write("%s\n" % AddUserConfGenerator._MODIFIED_MSG)
                        tmp.write("LAST_GID=%d \n" % self.gid_max)
                    elif l[0] == "LAST_UID" and l[1] != self.uid_max:
                        modified = True
                        tmp.write("%s\n" % AddUserConfGenerator._MODIFIED_MSG)
                        tmp.write("LAST_UID=%d \n" % self.uid_max)

        if modified:
            jetpack.util.update_file_permissions(tmp_filename, st.st_uid, st.st_gid, 0o644)
            shutil.move(tmp_filename, self.filename)

class AuthorizedKeysGenerator(jetpack.util.LockedFileManager):
    """ Manages an authorized keys file in a threadsafe way 
    
    keys is a dictionary of "identifier" -> [list of keys] mapping
    an identifier of None means that we will never remove keys from the file, whatever is in the list will simply be appended
    
    if there is an identifier, it will be inserted into the comment section of the key line. we will ensure that the authorized_keys
    file has "exactly" the list of keys in it with the given identifier. This means we will remove existing keys and add new keys within
    the same identifier group (this allows us to revoke/rotate keys).
    """

    def __init__(self, filename, keys, uid, gid):
        self.filename = filename
        self.authorized_keys = []
        self.uid = uid
        self.gid = gid
        self.keys = {}

        # Sanitize the keys passed in to ensure there are no newlines, fixes #5198646
        for identifier, key_list in keys.items():
           self.keys[identifier] = [key.strip() for key in key_list]

    def directory(self):
        return os.path.dirname(self.filename)

    def _read_authorized_keys(self):
        # Read in the authorized keys file and store it in keys
        self.authorized_keys = []
        if os.path.isfile(self.filename):
            with open(self.filename) as f:
                self.authorized_keys = f.read().splitlines()

    def check_for_update(self):
        self._read_authorized_keys()
        # Insert our keys into the authorized keys file and return if there is an update needed
        changes_detected = False
        for identifier, keys in self.keys.items():
            if identifier == '':
                for key in keys:
                    if key not in self.authorized_keys:
                        self.authorized_keys.append(key)
                        changes_detected = True
            else:
                managed_keys = ["%s %s" % (k, identifier) for k in keys]
                existing_keys = list(self.authorized_keys)  # Copy authorized keys

                # For all the keys we already have, check to see if it is part of our group if it is, see if it's been removed
                for k in existing_keys:
                    if k.endswith(identifier):
                        if k not in managed_keys:
                            self.authorized_keys.remove(k)
                            changes_detected = True

                # For every key we are managing check to see if we have to add it
                for k in managed_keys: 
                    if k not in self.authorized_keys:
                        self.authorized_keys.append(k)
                        changes_detected = True
        return changes_detected

    def write(self):
        tmp_filename = self.filename + ".tmp"
        with open(tmp_filename, 'w') as f:
            for key in self.authorized_keys:
                f.write("%s\n" % key)
        jetpack.util.update_file_permissions(tmp_filename, self.uid, self.gid, 0o644)
        shutil.move(tmp_filename, self.filename)

class SSHConfigGenerator(jetpack.util.LockedFileManager):
    """ Dead simple SSH Config file generator that only writes one specific config right now """
    def __init__(self, filename, uid, gid):
        self.filename = filename
        self.uid = uid
        self.gid = gid

    def directory(self):
        return os.path.dirname(self.filename)

    def check_for_update(self):
        return not os.path.isfile(self.filename)

    def write(self):
        content = """# Automatically generated by jetpackd - StrictHostKeyChecking is disabled to allow for passwordless SSH between Azure nodes
Host *
    StrictHostKeyChecking no
"""
        tmp_filename = self.filename + '.tmp'
        with open(tmp_filename, 'w') as f:
            f.write(content)
        jetpack.util.update_file_permissions(tmp_filename, self.uid, self.gid, 0o400)
        shutil.move(tmp_filename, self.filename)

class SSHKeypairGeneratorExecption(Exception):
    pass

class SSHKeypairGenerator(jetpack.util.LockedFileManager):
    """ Generates a private/public keypair in a threadsafe way 
    
    Usage:
    generator = SSHKeypairGenerator("/shared/home/test-user/.ssh/id_rsa", uid, gid)
    generator.update()

    Outcome:
    /shared/home/test-user/.ssh/id_rsa and /shared/home/test-user/.ssh/id_rsa.pub will exist
    """
    def __init__(self, private_key_path, uid, gid):
        self.private_key_path = private_key_path
        self.public_key_path = self.private_key_path + ".pub"
        self.uid = uid
        self.gid = gid
        self.private_key = None
        self.public_key = None
    
    def directory(self):
        return os.path.dirname(self.private_key_path)
    
    def check_for_update(self):
        """ We need to update the keys if either (or both) the keys are missing """
        if not os.path.isfile(self.private_key_path) or not os.path.isfile(self.public_key_path):
            return True
        else:
            return False

    def update(self):
        """ Generates the keypair and stores the keypair in memory """
        modified = super(SSHKeypairGenerator, self).update()
        
        with open(self.private_key_path) as f:
            self.private_key = f.read().strip()
        with open(self.public_key_path) as f:
            self.public_key = f.read().strip()
        
        return modified

    def write(self):
        tmp_private_key_path = self.private_key_path + ".tmp"
        tmp_public_key_path = tmp_private_key_path + ".pub"

        if not os.path.isfile(self.private_key_path):
            try:
                subprocess.check_output(['ssh-keygen', '-t', 'rsa', '-f', tmp_private_key_path, '-N', '', '-q'], stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                if e.output is not None:
                    e.output = e.output.strip()
                raise SSHKeypairGeneratorExecption("Failure to run `ssh-keygen`: %s (%s)" % (e.output, e.returncode))
            
            jetpack.util.update_file_permissions(tmp_private_key_path, self.uid, self.gid, 0o600)
            jetpack.util.update_file_permissions(tmp_public_key_path, self.uid, self.gid, 0o644)
            shutil.move(tmp_private_key_path, self.private_key_path)
            shutil.move(tmp_public_key_path, self.public_key_path)
        
        # If somehow we ended up without a public key (like the process died mid shutil.move we can regenerate the key)
        if not os.path.isfile(self.public_key_path):
            try:
                public_key = subprocess.check_output(['ssh-keygen', '-y', '-f', self.private_key_path])
                with open(tmp_public_key_path, 'w') as f:
                    f.write(public_key)
            except subprocess.CalledProcessError as e:
                if e.output is not None:
                    e.output = e.output.strip()
                raise SSHKeypairGeneratorExecption("Failure to run `ssh-keygen`: %s (%s)" % (e.output, e.returncode))

            jetpack.util.update_file_permissions(tmp_public_key_path, self.uid, self.gid, 0o644)
            shutil.move(tmp_public_key_path, self.public_key_path)
