""" Module that wraps the version of azcopy installed alongside Jetpack """
import logging
import json
import time
import pathlib
from typing import Tuple, List, Optional

from jetpack.util import JETPACK_USER_AGENT
from jetpack.util import UserError
import jetpack.lockers
import jetpack.util
import jetpack.util.command

logger = logging.getLogger(__name__)

# Some functions for turning the azcopy cli into a python API. This supports explicit Managed Identity authentication.
# Note: the API supports uploading data to a storage account, but the azcopy CLI only supports a single managed identity
# used for both ends of the transfer, and the design is to assign the "Storage Blob Data Reader" role for the auto-assigned identity.
# This means customers will need to use a non-locker based override for the identity to use when uploading.

class AzCopyError(UserError):
    pass


def copy(src_path: str, dst_path: str, src_locker: jetpack.lockers.Locker = None, dst_locker: jetpack.lockers.Locker = None, recursive: bool = False, client_id_override: Optional[str] = None) -> None:
    """ Runs `azcopy copy` from src->dst using the specified lockers (None means local filesystem)

    :param str src_path: The path to copy from
    :param src_locker: The locker to copy from (or None for local filesystem)
    :param str dst_path: The path to copy to
    :param dst_locker: The locker to copy to (or None for local filesystem) 
    """
    if src_locker is None and dst_locker is None:
        raise ValueError("At least one of src_locker and dst_locker must be specified")

    src, dst = _get_src_and_dst(src_path, dst_path, src_locker, dst_locker)
    args: List[str] = ['copy', src, dst]
    if recursive:
        args.append('--recursive')

    client_id = get_client_id(src_locker, dst_locker, client_id_override)
    _execute(args, client_id=client_id)


def sync(src_path: str, dst_path: str, src_locker: jetpack.lockers.Locker = None, dst_locker: jetpack.lockers.Locker = None, recursive: bool = False, delete_destination: bool = False, client_id_override: Optional[str] = None):
    """ Runs `azcopy sync` from src->dst using the specified lockers (None means local filesystem)

    :param str src_path: The path to sync from
    :param src_locker: The locker to sync from (or None for local filesystem)
    :param str dst_path: The path to sync to
    :param dst_locker: The locker to sync to (or None for local filesystem) 
    """

    if src_locker is None and dst_locker is None:
        raise ValueError("At least one of src_locker and dst_locker must be specified")

    src, dst = _get_src_and_dst(src_path, dst_path, src_locker, dst_locker)
    args: List[str] = ['sync', src, dst]
    if recursive:
        args.append('--recursive')
    if delete_destination:
        args.append('--delete-destination=true')

    client_id = get_client_id(src_locker, dst_locker, client_id_override)
    _execute(args, client_id=client_id)


def file_exists(src_path: str, src_locker: jetpack.lockers.Locker = None) -> bool:
    """ Runs `azcopy list` on src_path to check if the file exists

    :param str src_path: The path to test
    :param src_locker: The locker to test from
    """

    if src_locker is None:
        raise ValueError("src_locker must be specified")

    src, _ = _get_src_and_dst(src_path, None, src_locker, None)
    args: List[str] = ['list', src]

    client_id = get_client_id(src_locker, None)

    # this variable needs to be a reference value so the message_handler function can modify it
    # i.e. don't change this to a simple integer (non-reference value)
    tracker = {"file_count": 0}
    def message_handler(status):
        if status['MessageType'] == "ListObject":
            tracker["file_count"] += 1
            return False # false to stop further processing for this message
        return True

    _execute(args, client_id=client_id, message_handler=message_handler)
    return tracker["file_count"] > 0


def _get_src_and_dst(src_path: str, dst_path: str, src_locker: Optional[jetpack.lockers.Locker] = None, dst_locker: Optional[jetpack.lockers.Locker] = None) -> Tuple[str, str]:    
    """ Constructs src and dst paths/URLs """
    src: str = src_path
    if src_locker is not None:
        src = src_locker.url(src_path, include_sas_token=True)
    dst: str = dst_path
    if dst_locker is not None:
        dst = dst_locker.url(dst_path, include_sas_token=True)

    if jetpack.util.is_windows():
        # Windows binaries have to parse their own command lines, and it has historically been a bit quirky.
        # To avoid some of those quirks and their interaction with azcopy's directory/file detection we
        # simply replace backslashes with forward slashes when on windows.
        # See: https://learn.microsoft.com/en-us/cpp/cpp/main-function-command-line-args
        if src is not None:
            src = src.replace("\\", "/")
        if dst is not None:
            dst = dst.replace("\\", "/")

    return (src, dst)


def _azcopy() -> str:
    """ Returns the absolute path to azcopy """
    azcopy_bin: str = "azcopy"

    if jetpack.util.is_windows():
        azcopy_bin = 'azcopy.exe'
        return str(pathlib.PureWindowsPath(jetpack.util.jetpack_home(), 'system', 'embedded', 'bin', azcopy_bin))
    else:
        return str(pathlib.PurePosixPath(jetpack.util.jetpack_home(), 'system', 'embedded', 'bin', azcopy_bin))

def get_client_id(src_locker: jetpack.lockers.Locker, dst_locker: jetpack.lockers.Locker, client_id_override: Optional[str] = None) -> str:
    """ Returns the client id to use for the transfer, which cannot be both required and different for src and dst """

    if client_id_override is not None:
        return client_id_override
    
    src_client_id = src_locker.client_id if src_locker is not None else None
    dst_client_id = dst_locker.client_id if dst_locker is not None else None

    if src_client_id is None and dst_client_id is None:
        return None
    
    if src_client_id is not None and dst_client_id is None:
        # only the source needs a client id
        return src_client_id
    
    if dst_client_id is not None and src_client_id is None:
        # only the destination needs a client id
        return dst_client_id
    
    if src_client_id == dst_client_id:
        return src_client_id
        
    raise ValueError(f"Both lockers have different client ids: {src_client_id} and {dst_client_id}")

def _execute(args: List[str], client_id:Optional[str] = None, message_handler=None):
    cmd = [_azcopy()]
    cmd.extend(args)
    # Ensure we will get JSON output
    cmd.append('--output-type=json')

    custom_env = {
        'AZCOPY_USER_AGENT_PREFIX': JETPACK_USER_AGENT
    }

    if client_id:
        custom_env['AZCOPY_AUTO_LOGIN_TYPE'] = 'MSI'
        custom_env['AZCOPY_MSI_CLIENT_ID'] = client_id
    else:
        # without client id, we assume we have a URL with a SAS token in it
        pass

    start = time.time()

    try:
        with jetpack.util.command.begin("azcopy", cmd, env=custom_env) as c:
            logger.info("Running azcopy command: %s" % ' '.join(args))
            logger.info("Full azcopy command in %s" % c.script)
            for line in iter(c.proc.stdout.readline, ''):
                try:
                    status = json.loads(line.strip())
                    if message_handler is not None and not message_handler(status):
                        continue # skip further processing if the handler returns False

                    if status['MessageType'] == "Error":
                        # TODO: this should not really assume it is being used for cluster-init, but right now that is all it is used for
                        raise AzCopyError("Cluster-init error", description="AzCopy command failed", detail=f"[azcopy] {status['MessageContent']}", recommendation=c.get_recommendation())
                    elif status['MessageType'] == "Init":
                        pass
                    elif status['MessageType'] == "Info":
                        # TODO: Handle info messages
                        # NOTE: Sometimes errors are hidden in info messages (like malformed sas token)
                        logger.debug(status['MessageContent'])
                    elif status['MessageType'] == "Progress":
                        # TODO: We can read the % complete, IOPS, etc. and pass them to a callback
                        pass
                    elif status['MessageType'] == "EndOfJob":
                        logger.debug(status['MessageContent'])
                    else:
                        logger.warning("Unknown azcopy message type: '%s'" % status['MessageType'])
                except ValueError:  # JSON loading error
                    raise Exception("Unable to parse JSON output from azcopy")

        elapsed = time.time() - start
        if c.proc.returncode != 0:
            raise Exception("Unknown error running azcopy (return code %s, %f sec)" % (c.proc.returncode, elapsed))

        logger.info(f"azcopy command succeeded ({elapsed} sec)")
    except Exception:
        logger.exception("Unable to perform azcopy command: %s" % cmd)
        raise
