import unittest
from jetpack.admin import configure
from unittest import mock
import jetpack.util


class TestAdminConfigure(unittest.TestCase):
    def setUp(self):
        pass

    def test_init(self):
        c = configure.JetpackConfigure()
        self.assertEqual("", c.cyclecloud_bootstrap)
        self.assertEqual("", c.cyclecloud_home)

    def test_set_paths(self):

        # Windows + no enviornment variables
        with mock.patch.dict('os.environ', clear=True):
            with mock.patch('platform.system', mock.MagicMock(return_value="Windows")):
                c = configure.JetpackConfigure()
                c._set_paths()
                self.assertEqual("C:\\cycle\\jetpack", c.cyclecloud_home)
                self.assertEqual("C:\\cycle\\jetpack/system/bootstrap", c.cyclecloud_bootstrap)

            # Linux + no enviornment variables
            with mock.patch('platform.system', mock.MagicMock(return_value="Linux")):
                c = configure.JetpackConfigure()
                c._set_paths()
                self.assertEqual("/opt/cycle/jetpack", c.cyclecloud_home)
                self.assertEqual("/opt/cycle/jetpack/system/bootstrap", c.cyclecloud_bootstrap)

        # Linux with environment variables
        with mock.patch('platform.system', mock.MagicMock(return_value="Linux")):
            with mock.patch.dict('os.environ', {'CYCLECLOUD_HOME': '/tmp/cyclecloud',
                                                'CYCLECLOUD_BOOTSTRAP': '/tmp/bootstrap'}):
                c = configure.JetpackConfigure()
                c._set_paths()
                self.assertEqual("/tmp/cyclecloud", c.cyclecloud_home)
                self.assertEqual("/tmp/bootstrap", c.cyclecloud_bootstrap)

        # Windows with environment variables
        with mock.patch('platform.system', mock.MagicMock(return_value="Windows")):
            with mock.patch.dict('os.environ', {'CYCLECLOUD_HOME': 'C:\\cyclecloud',
                                                'CYCLECLOUD_BOOTSTRAP': 'C:\\bootstrap'}):
                c = configure.JetpackConfigure()
                c._set_paths()
                self.assertEqual("C:\\cyclecloud", c.cyclecloud_home)
                self.assertEqual("C:\\bootstrap", c.cyclecloud_bootstrap)

    @mock.patch('jetpack.config.get')
    @mock.patch('jetpack.converge.userdata.cache_userdata')
    def test_transform_userdata_to_dict(self, mock_cache_userdata, mock_get):
        c = configure.JetpackConfigure()

        # Sample userdata
        mock_cache_userdata.return_value = {
            'user_data': {
                'config': {
                    'webServer': 'http://127.0.0.1:37140',
                },
                'pool': {
                    'name': 'my-cluster-name'
                },
            }
        }

        def _get(v):
            data = {
                'cyclecloud.config.username': 'cyclecloud_access',
                'cyclecloud.config.password': 'YzkwLAJJNL',
                'cyclecloud.node.template': 'master',
                'cyclecloud.cluster.start_count': 5,
                'cyclecloud.cluster.session_id': '789'
            }
            return data[v]
        mock_get.side_effect = _get

        with mock.patch.object(jetpack.util, 'get_instance_id', return_value="i-abcd"):
            actual = c._transform_userdata_to_dict()
            expected = {
                'identity': {
                    'cluster_session_id': '789',
                    'cluster_name': 'my-cluster-name',
                    'instance_id': 'i-abcd',
                    'cluster_start_count': 5,
                    'template': 'master'
                },
                'cycle_server': {
                    'webserver': 'http://127.0.0.1:37140',
                    'api_username': 'cyclecloud_access',
                    'api_password': 'YzkwLAJJNL'
                }
            }
            self.assertEqual(actual, expected)


    def test_run(self):
        with mock.patch.object(configure.JetpackConfigure, '_set_paths', return_value=None) as mock_set_paths:
                with mock.patch.object(configure.JetpackConfigure, '_transform_userdata_to_dict') as mock_transform_userdata_to_dict:
                    mock_transform_userdata_to_dict.return_value = {'foo': 'bar'}
                    with mock.patch('jetpack.util.write_config') as mock_write_config:
                        c = configure.JetpackConfigure()
                        c.run('/tmp/user-data.json', '/tmp/jetpack.ini')
                        mock_set_paths.assert_called_once()
                        mock_transform_userdata_to_dict.assert_called_once_with()
                        mock_write_config.assert_called_once_with({'foo': 'bar'}, '/tmp/jetpack.ini')

    def test_execute(self):
        with mock.patch('jetpack.admin.configure.JetpackConfigure') as MockJetpackConfigure:
            configure.execute(user_data_path='/tmp/user-data.json', jetpack_config='/tmp/jetpack.ini')
            MockJetpackConfigure.return_value.run.assert_called_once_with(user_data_path='/tmp/user-data.json', jetpack_config='/tmp/jetpack.ini')
